@@ -48,6 +48,7 @@ typedef struct sgttyb conmode;
 # endif
 #elif defined _WIN32
 #include <winioctl.h>
+#include <conio.h>
 typedef DWORD conmode;
 
 #define LAST_ERROR rb_w32_map_errno(GetLastError())
@@ -74,6 +75,9 @@ getattr(int fd, conmode *t)
 #endif
 
 static ID id_getc, id_console, id_close, id_min, id_time;
+#if ENABLE_IO_GETPASS
+static ID id_gets;
+#endif
 
 #ifndef HAVE_RB_F_SEND
 static ID id___send__;
@@ -277,8 +281,7 @@ ttymode(VALUE io, VALUE (*func)(VALUE), void (*setter)(conmode *, void *), void
     }
     if (status) {
 	if (status == -1) {
-	    errno = error;
-	    rb_sys_fail(0);
+	    rb_syserr_fail(error, 0);
 	}
 	rb_jump_tag(status);
     }
@@ -295,6 +298,14 @@ ttymode(VALUE io, VALUE (*func)(VALUE), void (*setter)(conmode *, void *), void
  *
  * will read and return a line without echo back and line editing.
  *
+ * The parameter +min+ specifies the minimum number of bytes that
+ * should be received when a read operation is performed. (default: 1)
+ *
+ * The parameter +time+ specifies the timeout in _seconds_ with a
+ * precision of 1/10 of a second. (default: 0)
+ *
+ * Refer to the manual page of termios for further details.
+ *
  * You must require 'io/console' to use this method.
  */
 static VALUE
@@ -312,6 +323,8 @@ console_raw(int argc, VALUE *argv, VALUE io)
  *
  * If the terminal mode needs to be back, use io.raw { ... }.
  *
+ * See IO#raw for details on the parameters.
+ *
  * You must require 'io/console' to use this method.
  */
 static VALUE
@@ -373,11 +386,13 @@ console_set_cooked(VALUE io)
     return io;
 }
 
+#ifndef _WIN32
 static VALUE
 getc_call(VALUE io)
 {
-    return rb_funcall2(io, id_getc, 0, 0);
+    return rb_funcallv(io, id_getc, 0, 0);
 }
+#endif
 
 /*
  * call-seq:
@@ -385,13 +400,57 @@ getc_call(VALUE io)
  *
  * Reads and returns a character in raw mode.
  *
+ * See IO#raw for details on the parameters.
+ *
  * You must require 'io/console' to use this method.
  */
 static VALUE
 console_getch(int argc, VALUE *argv, VALUE io)
 {
     rawmode_arg_t opts, *optp = rawmode_opt(argc, argv, &opts);
+#ifndef _WIN32
     return ttymode(io, getc_call, set_rawmode, optp);
+#else
+    rb_io_t *fptr;
+    VALUE str;
+    wint_t c;
+    int w, len;
+    char buf[8];
+    struct timeval *to = NULL, tv;
+
+    GetOpenFile(io, fptr);
+    if (optp) {
+	if (optp->vtime) {
+	    to = &tv;
+	    tv.tv_sec = optp->vtime / 10;
+	    tv.tv_usec = (optp->vtime % 10) * 100000;
+	}
+	if (optp->vmin != 1) {
+	    rb_warning("min option ignored");
+	}
+    }
+    w = rb_wait_for_single_fd(fptr->fd, RB_WAITFD_IN, to);
+    if (w < 0) rb_eof_error();
+    if (!(w & RB_WAITFD_IN)) return Qnil;
+    c = _getwch();
+    switch (c) {
+      case WEOF:
+	return Qnil;
+      case 0x00:
+      case 0xe0:
+	buf[0] = (char)c;
+	c = _getwch();
+	len = 1;
+	do {
+	    buf[len++] = (unsigned char)c;
+	} while ((c >>= CHAR_BIT) && len < (int)sizeof(buf));
+	return rb_str_new(buf, len);
+      default:
+	len = rb_uv_to_utf8(buf, c);
+	str = rb_utf8_str_new(buf, len);
+	return rb_str_conv_enc(str, NULL, rb_default_external_encoding());
+    }
+#endif
 }
 
 /*
@@ -519,16 +578,23 @@ console_set_winsize(VALUE io, VALUE size)
 #if defined _WIN32
     HANDLE wh;
     int newrow, newcol;
+    BOOL ret;
 #endif
     VALUE row, col, xpixel, ypixel;
     const VALUE *sz;
     int fd;
+    long sizelen;
 
     GetOpenFile(io, fptr);
     size = rb_Array(size);
-    rb_check_arity(RARRAY_LENINT(size), 2, 4);
+    if ((sizelen = RARRAY_LEN(size)) != 2 && sizelen != 4) {
+	rb_raise(rb_eArgError,
+		 "wrong number of arguments (given %ld, expected 2 or 4)",
+		 sizelen);
+    }
     sz = RARRAY_CONST_PTR(size);
-    row = sz[0], col = sz[1], xpixel = sz[2], ypixel = sz[3];
+    row = sz[0], col = sz[1], xpixel = ypixel = Qnil;
+    if (sizelen == 4) xpixel = sz[2], ypixel = sz[3];
     fd = GetWriteFD(fptr);
 #if defined TIOCSWINSZ
     ws.ws_row = ws.ws_col = ws.ws_xpixel = ws.ws_ypixel = 0;
@@ -550,17 +616,21 @@ console_set_winsize(VALUE io, VALUE size)
     if (!GetConsoleScreenBufferInfo(wh, &ws)) {
 	rb_syserr_fail(LAST_ERROR, "GetConsoleScreenBufferInfo");
     }
-    if ((ws.dwSize.X < newcol && (ws.dwSize.X = newcol, 1)) ||
-	(ws.dwSize.Y < newrow && (ws.dwSize.Y = newrow, 1))) {
-	if (!SetConsoleScreenBufferSize(wh, ws.dwSize)) {
-	    rb_syserr_fail(LAST_ERROR, "SetConsoleScreenBufferInfo");
-	}
-    }
+    ws.dwSize.X = newcol;
+    ret = SetConsoleScreenBufferSize(wh, ws.dwSize);
     ws.srWindow.Left = 0;
     ws.srWindow.Top = 0;
-    ws.srWindow.Right = newcol;
-    ws.srWindow.Bottom = newrow;
-    if (!SetConsoleWindowInfo(wh, FALSE, &ws.srWindow)) {
+    ws.srWindow.Right = newcol-1;
+    ws.srWindow.Bottom = newrow-1;
+    if (!SetConsoleWindowInfo(wh, TRUE, &ws.srWindow)) {
+	rb_syserr_fail(LAST_ERROR, "SetConsoleWindowInfo");
+    }
+    /* retry when shrinking buffer after shrunk window */
+    if (!ret && !SetConsoleScreenBufferSize(wh, ws.dwSize)) {
+	rb_syserr_fail(LAST_ERROR, "SetConsoleScreenBufferInfo");
+    }
+    /* remove scrollbar if possible */
+    if (!SetConsoleWindowInfo(wh, TRUE, &ws.srWindow)) {
 	rb_syserr_fail(LAST_ERROR, "SetConsoleWindowInfo");
     }
 #endif
@@ -716,15 +786,17 @@ console_key_pressed_p(VALUE io, VALUE k)
     }
     else {
 	const struct vktable *t;
+	const char *kn;
 	if (SYMBOL_P(k)) {
 	    k = rb_sym2str(k);
+	    kn = RSTRING_PTR(k);
 	}
 	else {
-	    StringValueCStr(k);
+	    kn = StringValuePtr(k);
 	}
-	t = console_win32_vk(RSTRING_PTR(k), RSTRING_LEN(k));
+	t = console_win32_vk(kn, RSTRING_LEN(k));
 	if (!t || (vk = (short)t->vk) == -1) {
-	    rb_raise(rb_eArgError, "unknown virtual key code: %"PRIsVALUE, k);
+	    rb_raise(rb_eArgError, "unknown virtual key code: % "PRIsVALUE, k);
 	}
     }
     return GetKeyState(vk) & 0x80 ? Qtrue : Qfalse;
@@ -842,9 +914,83 @@ console_dev(int argc, VALUE *argv, VALUE klass)
 static VALUE
 io_getch(int argc, VALUE *argv, VALUE io)
 {
-    return rb_funcall2(io, id_getc, argc, argv);
+    return rb_funcallv(io, id_getc, argc, argv);
 }
 
+#if ENABLE_IO_GETPASS
+static VALUE
+puts_call(VALUE io)
+{
+    return rb_io_write(io, rb_default_rs);
+}
+
+static VALUE
+getpass_call(VALUE io)
+{
+    return ttymode(io, rb_io_gets, set_noecho, NULL);
+}
+
+static void
+prompt(int argc, VALUE *argv, VALUE io)
+{
+    if (argc > 0 && !NIL_P(argv[0])) {
+	VALUE str = argv[0];
+	StringValueCStr(str);
+	rb_check_safe_obj(str);
+	rb_io_write(io, str);
+    }
+}
+
+static VALUE
+str_chomp(VALUE str)
+{
+    if (!NIL_P(str)) {
+	str = rb_funcallv(str, rb_intern("chomp!"), 0, 0);
+    }
+    return str;
+}
+
+/*
+ * call-seq:
+ *   io.getpass(prompt=nil)       -> string
+ *
+ * Reads and returns a line without echo back.
+ * Prints +prompt+ unless it is +nil+.
+ *
+ * You must require 'io/console' to use this method.
+ */
+static VALUE
+console_getpass(int argc, VALUE *argv, VALUE io)
+{
+    VALUE str, wio;
+
+    rb_check_arity(argc, 0, 1);
+    wio = rb_io_get_write_io(io);
+    if (wio == io && io == rb_stdin) wio = rb_stderr;
+    prompt(argc, argv, wio);
+    str = rb_ensure(getpass_call, io, puts_call, wio);
+    return str_chomp(str);
+}
+
+/*
+ * call-seq:
+ *   io.getpass(prompt=nil)       -> string
+ *
+ * See IO#getpass.
+ */
+static VALUE
+io_getpass(int argc, VALUE *argv, VALUE io)
+{
+    VALUE str;
+
+    rb_check_arity(argc, 0, 1);
+    prompt(argc, argv, io);
+    str = str_chomp(rb_funcallv(io, id_gets, 0, 0));
+    puts_call(io);
+    return str;
+}
+#endif
+
 /*
  * IO console methods
  */
@@ -853,6 +999,9 @@ Init_console(void)
 {
 #undef rb_intern
     id_getc = rb_intern("getc");
+#if ENABLE_IO_GETPASS
+    id_gets = rb_intern("gets");
+#endif
     id_console = rb_intern("console");
     id_close = rb_intern("close");
     id_min = rb_intern("min");
@@ -884,9 +1033,15 @@ InitVM_console(void)
     rb_define_method(rb_cIO, "cursor", console_cursor_pos, 0);
     rb_define_method(rb_cIO, "cursor=", console_cursor_set, 1);
     rb_define_method(rb_cIO, "pressed?", console_key_pressed_p, 1);
+#if ENABLE_IO_GETPASS
+    rb_define_method(rb_cIO, "getpass", console_getpass, -1);
+#endif
     rb_define_singleton_method(rb_cIO, "console", console_dev, -1);
     {
 	VALUE mReadable = rb_define_module_under(rb_cIO, "generic_readable");
 	rb_define_method(mReadable, "getch", io_getch, -1);
+#if ENABLE_IO_GETPASS
+	rb_define_method(mReadable, "getpass", io_getpass, -1);
+#endif
     }
 }
