@@ -1,6 +1,6 @@
 # frozen_string_literal: true
 #
-# $Id: lexer.rb 60069 2017-09-30 06:42:24Z nobu $
+# $Id$
 #
 # Copyright (c) 2004,2005 Minero Aoki
 #
@@ -46,36 +46,53 @@ class Ripper
   end
 
   class Lexer < ::Ripper   #:nodoc: internal use only
-    Elem = Struct.new(:pos, :event, :tok, :state)
-    class Elem
-      class List < ::Array
-        def inspect
-          super.sub!(/\d+(?=\]\z)/, Ripper.lex_state_name(self[3]))
-        end
+    State = Struct.new(:to_int, :to_s) do
+      alias to_i to_int
+      def initialize(i) super(i, Ripper.lex_state_name(i)).freeze end
+      # def inspect; "#<#{self.class}: #{self}>" end
+      alias inspect to_s
+      def pretty_print(q) q.text(to_s) end
+      def ==(i) super or to_int == i end
+      def &(i) self.class.new(to_int & i) end
+      def |(i) self.class.new(to_int & i) end
+      def allbits?(i) to_int.allbits?(i) end
+      def anybits?(i) to_int.anybits?(i) end
+      def nobits?(i) to_int.nobits?(i) end
+    end
 
-        def pretty_print(q) # :nodoc:
-          pos, event, tok, state = self
-          q.group(1, '[', ']') {
-            q.pp pos
-            q.comma_breakable
-            q.pp event
-            q.comma_breakable
-            q.pp tok
-            q.comma_breakable
-            q.text(Ripper.lex_state_name(state))
-          }
-        end
+    Elem = Struct.new(:pos, :event, :tok, :state, :message) do
+      def initialize(pos, event, tok, state, message = nil)
+        super(pos, event, tok, State.new(state), message)
+      end
 
-        def pretty_print_cycle(q) # :nodoc:
-          q.text('[...]')
-        end
+      def inspect
+        "#<#{self.class}: #{event}@#{pos[0]}:#{pos[1]}:#{state}: #{tok.inspect}#{": " if message}#{message}>"
+      end
+
+      def pretty_print(q)
+        q.group(2, "#<#{self.class}:", ">") {
+          q.breakable
+          q.text("#{event}@#{pos[0]}:#{pos[1]}")
+          q.breakable
+          q.text("token: ")
+          tok.pretty_print(q)
+          if message
+            q.breakable
+            q.text("message: ")
+            q.text(message)
+          end
+        }
       end
 
       def to_a
-        List[*values]
+        a = super
+        a.pop unless a.last
+        a
       end
     end
 
+    attr_reader :errors
+
     def tokenize
       parse().sort_by(&:pos).map(&:tok)
     end
@@ -84,7 +101,13 @@ class Ripper
       parse().sort_by(&:pos).map(&:to_a)
     end
 
+    # parse the code and returns elements including errors.
+    def scan
+      (parse() + errors + @stack.flatten).uniq.sort_by {|e| [*e.pos, (e.message ? -1 : 0)]}
+    end
+
     def parse
+      @errors = []
       @buf = []
       @stack = []
       super
@@ -94,11 +117,17 @@ class Ripper
 
     private
 
+    unless SCANNER_EVENT_TABLE.key?(:ignored_sp)
+      SCANNER_EVENT_TABLE[:ignored_sp] = 1
+      SCANNER_EVENTS << :ignored_sp
+      EVENTS << :ignored_sp
+    end
+
     def on_heredoc_dedent(v, w)
       ignored_sp = []
       heredoc = @buf.last
       heredoc.each_with_index do |e, i|
-        if Elem === e and e.event == :on_tstring_content
+        if Elem === e and e.event == :on_tstring_content and e.pos[1].zero?
           tok = e.tok.dup if w > 0 and /\A\s/ =~ e.tok
           if (n = dedent_string(e.tok, w)) > 0
             if e.tok.empty?
@@ -134,6 +163,12 @@ class Ripper
       @buf.push Elem.new([lineno(), column()], __callee__, tok, state())
     end
 
+    def on_error(mesg)
+      @errors.push Elem.new([lineno(), column()], __callee__, token(), state(), mesg)
+    end
+    alias on_parse_error on_error
+    alias compile_error on_error
+
     (SCANNER_EVENTS.map {|event|:"on_#{event}"} - private_instance_methods(false)).each do |event|
       alias_method event, :_push_token
     end
@@ -197,7 +232,7 @@ class Ripper
       if m = /[^\w\s$()\[\]{}?*+\.]/.match(pattern)
         raise CompileError, "invalid char in pattern: #{m[0].inspect}"
       end
-      buf = ''
+      buf = +''
       pattern.scan(/(?:\w+|\$\(|[()\[\]\{\}?*+\.]+)/) do |tok|
         case tok
         when /\w/
@@ -218,14 +253,14 @@ class Ripper
     end
 
     def map_tokens(tokens)
-      tokens.map {|pos,type,str| map_token(type.to_s.sub(/\Aon_/,'')) }.join
+      tokens.map {|pos,type,str| map_token(type.to_s.delete_prefix('on_')) }.join
     end
 
     MAP = {}
     seed = ('a'..'z').to_a + ('A'..'Z').to_a + ('0'..'9').to_a
     SCANNER_EVENT_TABLE.each do |ev, |
       raise CompileError, "[RIPPER FATAL] too many system token" if seed.empty?
-      MAP[ev.to_s.sub(/\Aon_/,'')] = seed.shift
+      MAP[ev.to_s.delete_prefix('on_')] = seed.shift
     end
 
     def map_token(tok)
