@@ -2,21 +2,26 @@
 
   inits.c -
 
-  $Author: nobu $
+  $Author$
   created at: Tue Dec 28 16:01:58 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
+#include "internal.h"
 
 #define CALL(n) {void Init_##n(void); Init_##n();}
 
 void
 rb_call_inits(void)
 {
-    CALL(RandomSeed);
+#if USE_TRANSIENT_HEAP
+    CALL(TransientHeap);
+#endif
+    CALL(vm_postponed_job);
+    CALL(Method);
+    CALL(RandomSeedCore);
     CALL(sym);
     CALL(var_tables);
     CALL(Object);
@@ -45,7 +50,6 @@ rb_call_inits(void)
     CALL(Time);
     CALL(Random);
     CALL(signal);
-    CALL(process);
     CALL(load);
     CALL(Proc);
     CALL(Binding);
@@ -55,9 +59,14 @@ rb_call_inits(void)
     CALL(VM);
     CALL(ISeq);
     CALL(Thread);
+    CALL(process);
     CALL(Cont);
     CALL(Rational);
     CALL(Complex);
     CALL(version);
+    CALL(vm_trace);
+    CALL(vm_stack_canary);
+    CALL(ast);
+    CALL(gc_stress);
 }
 #undef CALL
