@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 require 'tempfile'
 require 'rubygems'
 require 'rubygems/remote_fetcher'
@@ -12,6 +13,13 @@ require 'rubygems/remote_fetcher'
 #   @fetcher.data['http://gems.example.com/yaml'] = source_index.to_yaml
 #   Gem::RemoteFetcher.fetcher = @fetcher
 #
+#   use nested array if multiple response is needed
+#
+#   @fetcher.data['http://gems.example.com/sequence'] = [['Success', 200, 'OK'], ['Failed', 401, 'Unauthorized']]
+#
+#   @fetcher.fetch_path('http://gems.example.com/sequence') # => ['Success', 200, 'OK']
+#   @fetcher.fetch_path('http://gems.example.com/sequence') # => ['Failed', 401, 'Unauthorized']
+#
 #   # invoke RubyGems code
 #
 #   paths = @fetcher.paths
@@ -24,23 +32,17 @@ class Gem::FakeFetcher
 
   attr_reader :data
   attr_reader :last_request
-  attr_reader :api_endpoints
   attr_accessor :paths
 
   def initialize
     @data = {}
     @paths = []
-    @api_endpoints = {}
   end
 
-  def api_endpoint(uri)
-    @api_endpoints[uri] || uri
-  end
-
-  def find_data(path)
+  def find_data(path, nargs = 3)
     return File.read path.path if URI === path and 'file' == path.scheme
 
-    if URI === path and "URI::#{path.scheme.upcase}" != path.class.name then
+    if URI === path and "URI::#{path.scheme.upcase}" != path.class.name
       raise ArgumentError,
         "mismatch for scheme #{path.scheme} and class #{path.class}"
     end
@@ -49,28 +51,30 @@ class Gem::FakeFetcher
     @paths << path
     raise ArgumentError, 'need full URI' unless path =~ %r'^https?://'
 
-    unless @data.key? path then
+    unless @data.key? path
       raise Gem::RemoteFetcher::FetchError.new("no data for #{path}", path)
     end
 
-    @data[path]
+    data = @data[path]
+
+    data.flatten! and return data.shift(nargs) if data.respond_to?(:flatten!)
+    data
   end
 
-  def fetch_path path, mtime = nil, head = false
+  def fetch_path(path, mtime = nil, head = false)
     data = find_data(path)
 
-    if data.respond_to?(:call) then
+    if data.respond_to?(:call)
       data.call
     else
-      if path.to_s =~ /gz$/ and not data.nil? and not data.empty? then
-        data = Gem.gunzip data
+      if path.to_s =~ /gz$/ and not data.nil? and not data.empty?
+        data = Gem::Util.gunzip data
       end
-
       data
     end
   end
 
-  def cache_update_path uri, path = nil, update = true
+  def cache_update_path(uri, path = nil, update = true)
     if data = fetch_path(uri)
       open(path, 'wb') { |io| io.write data } if path and update
       data
@@ -92,7 +96,7 @@ class Gem::FakeFetcher
 
   def request(uri, request_class, last_modified = nil)
     data = find_data(uri)
-    body, code, msg = data
+    body, code, msg = (data.respond_to?(:call) ? data.call : data)
 
     @last_request = request_class.new uri.request_uri
     yield @last_request if block_given?
@@ -103,21 +107,13 @@ class Gem::FakeFetcher
     response
   end
 
-  def pretty_print q # :nodoc:
+  def pretty_print(q) # :nodoc:
     q.group 2, '[FakeFetcher', ']' do
       q.breakable
       q.text 'URIs:'
 
       q.breakable
       q.pp @data.keys
-
-      unless @api_endpoints.empty? then
-        q.breakable
-        q.text 'API endpoints:'
-
-        q.breakable
-        q.pp @api_endpoints.keys
-      end
     end
   end
 
@@ -127,7 +123,7 @@ class Gem::FakeFetcher
 
     raise ArgumentError, 'need full URI' unless path =~ %r'^http://'
 
-    unless @data.key? path then
+    unless @data.key? path
       raise Gem::RemoteFetcher::FetchError.new("no data for #{path}", path)
     end
 
@@ -136,9 +132,9 @@ class Gem::FakeFetcher
     data.respond_to?(:call) ? data.call : data.length
   end
 
-  def download spec, source_uri, install_dir = Gem.dir
+  def download(spec, source_uri, install_dir = Gem.dir)
     name = File.basename spec.cache_file
-    path = if Dir.pwd == install_dir then # see fetch_command
+    path = if Dir.pwd == install_dir  # see fetch_command
              install_dir
            else
              File.join install_dir, "cache"
@@ -146,7 +142,7 @@ class Gem::FakeFetcher
 
     path = File.join path, name
 
-    if source_uri =~ /^http/ then
+    if source_uri =~ /^http/
       File.open(path, "wb") do |f|
         f.write fetch_path(File.join(source_uri, "gems", name))
       end
@@ -157,7 +153,7 @@ class Gem::FakeFetcher
     path
   end
 
-  def download_to_cache dependency
+  def download_to_cache(dependency)
     found, _ = Gem::SpecFetcher.fetcher.spec_for_dependency dependency
 
     return if found.empty?
@@ -200,7 +196,7 @@ class Gem::TestCase::SpecFetcherSetup
   # Executes a SpecFetcher setup block.  Yields an instance then creates the
   # gems and specifications defined in the instance.
 
-  def self.declare test, repository
+  def self.declare(test, repository)
     setup = new test, repository
 
     yield setup
@@ -208,7 +204,7 @@ class Gem::TestCase::SpecFetcherSetup
     setup.execute
   end
 
-  def initialize test, repository # :nodoc:
+  def initialize(test, repository) # :nodoc:
     @test       = test
     @repository = repository
 
@@ -276,7 +272,7 @@ class Gem::TestCase::SpecFetcherSetup
   # The specification will be yielded before gem creation for customization,
   # but only the block or the dependencies may be set, not both.
 
-  def gem name, version, dependencies = nil, &block
+  def gem(name, version, dependencies = nil, &block)
     @operations << [:gem, name, version, dependencies, block]
   end
 
@@ -287,7 +283,7 @@ class Gem::TestCase::SpecFetcherSetup
   # The specification will be yielded before gem creation for customization,
   # but only the block or the dependencies may be set, not both.
 
-  def download name, version, dependencies = nil, &block
+  def download(name, version, dependencies = nil, &block)
     @operations << [:download, name, version, dependencies, block]
   end
 
@@ -306,7 +302,7 @@ class Gem::TestCase::SpecFetcherSetup
     require 'socket'
     require 'rubygems/remote_fetcher'
 
-    unless @test.fetcher then
+    unless @test.fetcher
       @test.fetcher = Gem::FakeFetcher.new
       Gem::RemoteFetcher.fetcher = @test.fetcher
     end
@@ -340,12 +336,12 @@ class Gem::TestCase::SpecFetcherSetup
   # The specification will be yielded before creation for customization,
   # but only the block or the dependencies may be set, not both.
 
-  def spec name, version, dependencies = nil, &block
+  def spec(name, version, dependencies = nil, &block)
     @operations << [:spec, name, version, dependencies, block]
   end
 
-  def write_spec spec # :nodoc:
-    open spec.spec_file, 'w' do |io|
+  def write_spec(spec) # :nodoc:
+    File.open spec.spec_file, 'w' do |io|
       io.write spec.to_ruby_for_cache
     end
   end
@@ -379,5 +375,5 @@ class TempIO < Tempfile
     flush
     Gem.read_binary path
   end
-end
 
+end
