@@ -1,12 +1,12 @@
+# frozen_string_literal: true
 #
 # tempfile - manipulates temporary files
 #
-# $Id: tempfile.rb 31723 2011-05-24 22:58:18Z nobu $
+# $Id$
 #
 
 require 'delegate'
 require 'tmpdir'
-require 'thread'
 
 # A utility class for managing temporary files. When you create a Tempfile
 # object, it will create a temporary file with a unique filename. A Tempfile
@@ -39,7 +39,7 @@ require 'thread'
 # that's it's unnecessary to explicitly delete a Tempfile after use, though
 # it's good practice to do so: not explicitly deleting unused Tempfiles can
 # potentially leave behind large amounts of tempfiles on the filesystem
-# until they're garbage collected. The existance of these temp files can make
+# until they're garbage collected. The existence of these temp files can make
 # it harder to determine a new Tempfile filename.
 #
 # Therefore, one should always call #unlink or close in an ensure block, like
@@ -47,7 +47,7 @@ require 'thread'
 #
 #   file = Tempfile.new('foo')
 #   begin
-#      ...do something with file...
+#      # ...do something with file...
 #   ensure
 #      file.close
 #      file.unlink   # deletes the temp file
@@ -79,12 +79,6 @@ require 'thread'
 # same Tempfile object from multiple threads then you should protect it with a
 # mutex.
 class Tempfile < DelegateClass(File)
-  MAX_TRY = 10  # :nodoc:
-  include Dir::Tmpname
-
-  # call-seq:
-  #    new(basename, [tmpdir = Dir.tmpdir], [options])
-  #
   # Creates a temporary file with permissions 0600 (= only readable and
   # writable by the owner) and opens it with mode "w+".
   #
@@ -117,54 +111,43 @@ class Tempfile < DelegateClass(File)
   # +File.open+. This is mostly useful for specifying encoding
   # options, e.g.:
   #
-  #   Tempfile.new('hello', '/home/aisaka', :encoding => 'ascii-8bit')
+  #   Tempfile.new('hello', '/home/aisaka', encoding: 'ascii-8bit')
   #
   #   # You can also omit the 'tmpdir' parameter:
-  #   Tempfile.new('hello', :encoding => 'ascii-8bit')
+  #   Tempfile.new('hello', encoding: 'ascii-8bit')
+  #
+  # Note: +mode+ keyword argument, as accepted by Tempfile, can only be
+  # numeric, combination of the modes defined in File::Constants.
   #
   # === Exceptions
   #
   # If Tempfile.new cannot find a unique filename within a limited
   # number of tries, then it will raise an exception.
-  def initialize(basename, *rest)
-    @data = []
-    @clean_proc = Remover.new(@data)
-    ObjectSpace.define_finalizer(self, @clean_proc)
+  def initialize(basename="", tmpdir=nil, mode: 0, **options)
+    warn "Tempfile.new doesn't call the given block.", uplevel: 1 if block_given?
 
-    create(basename, *rest) do |tmpname, n, opts|
-      mode = File::RDWR|File::CREAT|File::EXCL
-      perm = 0600
-      if opts
-        mode |= opts.delete(:mode) || 0
-        opts[:perm] = perm
-        perm = nil
-      else
-        opts = perm
-      end
-      self.class.locking(tmpname) do
-        @data[1] = @tmpfile = File.open(tmpname, mode, opts)
-        @data[0] = @tmpname = tmpname
-      end
-      @mode = mode & ~(File::CREAT|File::EXCL)
-      perm or opts.freeze
-      @opts = opts
+    @unlinked = false
+    @mode = mode|File::RDWR|File::CREAT|File::EXCL
+    ::Dir::Tmpname.create(basename, tmpdir, options) do |tmpname, n, opts|
+      opts[:perm] = 0600
+      @tmpfile = File.open(tmpname, @mode, opts)
+      @opts = opts.freeze
     end
+    ObjectSpace.define_finalizer(self, Remover.new(@tmpfile))
 
     super(@tmpfile)
   end
 
   # Opens or reopens the file with mode "r+".
   def open
-    @tmpfile.close if @tmpfile
-    @tmpfile = File.open(@tmpname, @mode, @opts)
-    @data[1] = @tmpfile
+    _close
+    mode = @mode & ~(File::CREAT|File::EXCL)
+    @tmpfile = File.open(@tmpfile.path, mode, @opts)
     __setobj__(@tmpfile)
   end
 
   def _close    # :nodoc:
-    @tmpfile.close if @tmpfile
-    @tmpfile = nil
-    @data[1] = nil if @data
+    @tmpfile.close
   end
   protected :_close
 
@@ -175,19 +158,14 @@ class Tempfile < DelegateClass(File)
   # If you don't explicitly unlink the temporary file, the removal
   # will be delayed until the object is finalized.
   def close(unlink_now=false)
-    if unlink_now
-      close!
-    else
-      _close
-    end
+    _close
+    unlink if unlink_now
   end
 
   # Closes and unlinks (deletes) the file. Has the same effect as called
   # <tt>close(true)</tt>.
   def close!
-    _close
-    unlink
-    ObjectSpace.undefine_finalizer(self)
+    close(true)
   end
 
   # Unlinks (deletes) the file from the filesystem. One should always unlink
@@ -196,7 +174,7 @@ class Tempfile < DelegateClass(File)
   #
   #   file = Tempfile.new('foo')
   #   begin
-  #      ...do something with file...
+  #      # ...do something with file...
   #   ensure
   #      file.close
   #      file.unlink   # deletes the temp file
@@ -217,94 +195,95 @@ class Tempfile < DelegateClass(File)
   #   file = Tempfile.new('foo')
   #   file.unlink   # On Windows this silently fails.
   #   begin
-  #      ... do something with file ...
+  #      # ... do something with file ...
   #   ensure
   #      file.close!   # Closes the file handle. If the file wasn't unlinked
   #                    # because #unlink failed, then this method will attempt
   #                    # to do so again.
   #   end
   def unlink
-    # keep this order for thread safeness
-    return unless @tmpname
+    return if @unlinked
     begin
-      if File.exist?(@tmpname)
-        File.unlink(@tmpname)
-      end
-      # remove tmpname from remover
-      @data[0] = @data[2] = nil
-      @tmpname = nil
+      File.unlink(@tmpfile.path)
+    rescue Errno::ENOENT
     rescue Errno::EACCES
       # may not be able to unlink on Windows; just ignore
+      return
     end
+    ObjectSpace.undefine_finalizer(self)
+    @unlinked = true
   end
   alias delete unlink
 
   # Returns the full path name of the temporary file.
   # This will be nil if #unlink has been called.
   def path
-    @tmpname
+    @unlinked ? nil : @tmpfile.path
   end
 
   # Returns the size of the temporary file.  As a side effect, the IO
   # buffer is flushed before determining the size.
   def size
-    if @tmpfile
-      @tmpfile.flush
-      @tmpfile.stat.size
-    elsif @tmpname
-      File.size(@tmpname)
+    if !@tmpfile.closed?
+      @tmpfile.size # File#size calls rb_io_flush_raw()
     else
-      0
+      File.size(@tmpfile.path)
     end
   end
   alias length size
 
   # :stopdoc:
-  class Remover
-    def initialize(data)
-      @pid = $$
-      @data = data
+  def inspect
+    if closed?
+      "#<#{self.class}:#{path} (closed)>"
+    else
+      "#<#{self.class}:#{path}>"
+    end
+  end
+
+  class Remover # :nodoc:
+    def initialize(tmpfile)
+      @pid = Process.pid
+      @tmpfile = tmpfile
     end
 
     def call(*args)
-      if @pid == $$
-        path, tmpfile = *@data
+      return if @pid != Process.pid
 
-        STDERR.print "removing ", path, "..." if $DEBUG
+      $stderr.puts "removing #{@tmpfile.path}..." if $DEBUG
 
-        tmpfile.close if tmpfile
-
-        # keep this order for thread safeness
-        if path
-          File.unlink(path) if File.exist?(path)
-        end
-
-        STDERR.print "done\n" if $DEBUG
+      @tmpfile.close
+      begin
+        File.unlink(@tmpfile.path)
+      rescue Errno::ENOENT
       end
+
+      $stderr.puts "done" if $DEBUG
     end
   end
-  # :startdoc:
 
   class << self
+    # :startdoc:
+
     # Creates a new Tempfile.
     #
     # If no block is given, this is a synonym for Tempfile.new.
     #
     # If a block is given, then a Tempfile object will be constructed,
     # and the block is run with said object as argument. The Tempfile
-    # oject will be automatically closed after the block terminates.
+    # object will be automatically closed after the block terminates.
     # The call returns the value of the block.
     #
-    # In any case, all arguments (+*args+) will be passed to Tempfile.new.
+    # In any case, all arguments (<code>*args</code>) will be passed to Tempfile.new.
     #
     #   Tempfile.open('foo', '/home/temp') do |f|
-    #      ... do something with f ...
+    #      # ... do something with f ...
     #   end
     #
     #   # Equivalent:
     #   f = Tempfile.open('foo', '/home/temp')
     #   begin
-    #      ... do something with f ...
+    #      # ... do something with f ...
     #   ensure
     #      f.close
     #   end
@@ -321,35 +300,55 @@ class Tempfile < DelegateClass(File)
         tempfile
       end
     end
-
-    # :stopdoc:
-
-    # yields with locking for +tmpname+ and returns the result of the
-    # block.
-    def locking(tmpname)
-      lock = tmpname + '.lock'
-      mkdir(lock)
-      yield
-    ensure
-      rmdir(lock) if lock
-    end
-
-    def mkdir(*args)
-      Dir.mkdir(*args)
-    end
-
-    def rmdir(*args)
-      Dir.rmdir(*args)
-    end
   end
 end
 
-if __FILE__ == $0
-#  $DEBUG = true
-  f = Tempfile.new("foo")
-  f.print("foo\n")
-  f.close
-  f.open
-  p f.gets # => "foo\n"
-  f.close!
+# Creates a temporary file as usual File object (not Tempfile).
+# It doesn't use finalizer and delegation.
+#
+# If no block is given, this is similar to Tempfile.new except
+# creating File instead of Tempfile.
+# The created file is not removed automatically.
+# You should use File.unlink to remove it.
+#
+# If a block is given, then a File object will be constructed,
+# and the block is invoked with the object as the argument.
+# The File object will be automatically closed and
+# the temporary file is removed after the block terminates.
+# The call returns the value of the block.
+#
+# In any case, all arguments (+basename+, +tmpdir+, +mode+, and
+# <code>**options</code>) will be treated as Tempfile.new.
+#
+#   Tempfile.create('foo', '/home/temp') do |f|
+#      # ... do something with f ...
+#   end
+#
+def Tempfile.create(basename="", tmpdir=nil, mode: 0, **options)
+  tmpfile = nil
+  Dir::Tmpname.create(basename, tmpdir, options) do |tmpname, n, opts|
+    mode |= File::RDWR|File::CREAT|File::EXCL
+    opts[:perm] = 0600
+    tmpfile = File.open(tmpname, mode, opts)
+  end
+  if block_given?
+    begin
+      yield tmpfile
+    ensure
+      unless tmpfile.closed?
+        if File.identical?(tmpfile, tmpfile.path)
+          unlinked = File.unlink tmpfile.path rescue nil
+        end
+        tmpfile.close
+      end
+      unless unlinked
+        begin
+          File.unlink tmpfile.path
+        rescue Errno::ENOENT
+        end
+      end
+    end
+  else
+    tmpfile
+  end
 end
