@@ -1,8 +1,13 @@
 # frozen_string_literal: true
-require "spec_helper"
 
 RSpec.describe "bundle platform" do
   context "without flags" do
+    let(:bundle_platform_platforms_string) do
+      platforms = [rb]
+      platforms.unshift(specific_local_platform) if Bundler.feature_flag.bundler_3_mode?
+      platforms.map {|pl| "* #{pl}" }.join("\n")
+    end
+
     it "returns all the output" do
       gemfile <<-G
         source "file://#{gem_repo1}"
@@ -17,7 +22,7 @@ RSpec.describe "bundle platform" do
 Your platform is: #{RUBY_PLATFORM}
 
 Your app has gems that work on these platforms:
-* ruby
+#{bundle_platform_platforms_string}
 
 Your Gemfile specifies a Ruby version requirement:
 * ruby #{RUBY_VERSION}
@@ -40,7 +45,7 @@ G
 Your platform is: #{RUBY_PLATFORM}
 
 Your app has gems that work on these platforms:
-* ruby
+#{bundle_platform_platforms_string}
 
 Your Gemfile specifies a Ruby version requirement:
 * ruby #{RUBY_VERSION}p#{RUBY_PATCHLEVEL}
@@ -61,7 +66,7 @@ G
 Your platform is: #{RUBY_PLATFORM}
 
 Your app has gems that work on these platforms:
-* ruby
+#{bundle_platform_platforms_string}
 
 Your Gemfile does not specify a Ruby version requirement.
 G
@@ -81,7 +86,7 @@ G
 Your platform is: #{RUBY_PLATFORM}
 
 Your app has gems that work on these platforms:
-* ruby
+#{bundle_platform_platforms_string}
 
 Your Gemfile specifies a Ruby version requirement:
 * ruby #{not_local_ruby_version}
@@ -144,6 +149,19 @@ G
       expect(out).to eq("ruby 1.8.7 (rbx 1.2.4)")
     end
 
+    it "handles truffleruby" do
+      gemfile <<-G
+        source "file://#{gem_repo1}"
+        ruby "2.5.1", :engine => 'truffleruby', :engine_version => '1.0.0-rc6'
+
+        gem "foo"
+      G
+
+      bundle "platform --ruby"
+
+      expect(out).to eq("ruby 2.5.1 (truffleruby 1.0.0-rc6)")
+    end
+
     it "raises an error if engine is used but engine version is not" do
       gemfile <<-G
         source "file://#{gem_repo1}"
@@ -250,27 +268,27 @@ G
 
   def should_be_ruby_version_incorrect
     expect(exitstatus).to eq(18) if exitstatus
-    expect(out).to be_include("Your Ruby version is #{RUBY_VERSION}, but your Gemfile specified #{not_local_ruby_version}")
+    expect(err).to be_include("Your Ruby version is #{RUBY_VERSION}, but your Gemfile specified #{not_local_ruby_version}")
   end
 
   def should_be_engine_incorrect
     expect(exitstatus).to eq(18) if exitstatus
-    expect(out).to be_include("Your Ruby engine is #{local_ruby_engine}, but your Gemfile specified #{not_local_tag}")
+    expect(err).to be_include("Your Ruby engine is #{local_ruby_engine}, but your Gemfile specified #{not_local_tag}")
   end
 
   def should_be_engine_version_incorrect
     expect(exitstatus).to eq(18) if exitstatus
-    expect(out).to be_include("Your #{local_ruby_engine} version is #{local_engine_version}, but your Gemfile specified #{local_ruby_engine} #{not_local_engine_version}")
+    expect(err).to be_include("Your #{local_ruby_engine} version is #{local_engine_version}, but your Gemfile specified #{local_ruby_engine} #{not_local_engine_version}")
   end
 
   def should_be_patchlevel_incorrect
     expect(exitstatus).to eq(18) if exitstatus
-    expect(out).to be_include("Your Ruby patchlevel is #{RUBY_PATCHLEVEL}, but your Gemfile specified #{not_local_patchlevel}")
+    expect(err).to be_include("Your Ruby patchlevel is #{RUBY_PATCHLEVEL}, but your Gemfile specified #{not_local_patchlevel}")
   end
 
   def should_be_patchlevel_fixnum
     expect(exitstatus).to eq(18) if exitstatus
-    expect(out).to be_include("The Ruby patchlevel in your Gemfile must be a string")
+    expect(err).to be_include("The Ruby patchlevel in your Gemfile must be a string")
   end
 
   context "bundle install" do
@@ -493,7 +511,7 @@ G
         build_gem "activesupport", "3.0"
       end
 
-      bundle "update"
+      bundle "update", :all => true
       expect(the_bundle).to include_gems "rack 1.2", "rack-obama 1.0", "activesupport 3.0"
     end
 
@@ -510,7 +528,7 @@ G
           build_gem "activesupport", "3.0"
         end
 
-        bundle "update"
+        bundle "update", :all => true
         expect(the_bundle).to include_gems "rack 1.2", "rack-obama 1.0", "activesupport 3.0"
       end
     end
@@ -527,7 +545,7 @@ G
         build_gem "activesupport", "3.0"
       end
 
-      bundle :update
+      bundle :update, :all => true
       should_be_ruby_version_incorrect
     end
 
@@ -543,7 +561,7 @@ G
         build_gem "activesupport", "3.0"
       end
 
-      bundle :update
+      bundle :update, :all => true
       should_be_engine_incorrect
     end
 
@@ -560,7 +578,7 @@ G
           build_gem "activesupport", "3.0"
         end
 
-        bundle :update
+        bundle :update, :all => true
         should_be_engine_version_incorrect
       end
     end
@@ -576,12 +594,12 @@ G
         build_gem "activesupport", "3.0"
       end
 
-      bundle :update
+      bundle :update, :all => true
       should_be_patchlevel_incorrect
     end
   end
 
-  context "bundle show" do
+  context "bundle info" do
     before do
       install_gemfile <<-G
         source "file://#{gem_repo1}"
@@ -590,32 +608,32 @@ G
     end
 
     it "prints path if ruby version is correct" do
-      gemfile <<-G
+      install_gemfile! <<-G
         source "file://#{gem_repo1}"
         gem "rails"
 
         #{ruby_version_correct}
       G
 
-      bundle "show rails"
+      bundle "info rails --path"
       expect(out).to eq(default_bundle_path("gems", "rails-2.3.2").to_s)
     end
 
     it "prints path if ruby version is correct for any engine" do
       simulate_ruby_engine "jruby" do
-        gemfile <<-G
+        install_gemfile! <<-G
           source "file://#{gem_repo1}"
           gem "rails"
 
           #{ruby_version_correct_engineless}
         G
 
-        bundle "show rails"
+        bundle "info rails --path"
         expect(out).to eq(default_bundle_path("gems", "rails-2.3.2").to_s)
       end
     end
 
-    it "fails if ruby version doesn't match" do
+    it "fails if ruby version doesn't match", :bundler => "< 3" do
       gemfile <<-G
         source "file://#{gem_repo1}"
         gem "rails"
@@ -627,7 +645,7 @@ G
       should_be_ruby_version_incorrect
     end
 
-    it "fails if engine doesn't match" do
+    it "fails if engine doesn't match", :bundler => "< 3" do
       gemfile <<-G
         source "file://#{gem_repo1}"
         gem "rails"
@@ -639,7 +657,7 @@ G
       should_be_engine_incorrect
     end
 
-    it "fails if engine version doesn't match" do
+    it "fails if engine version doesn't match", :bundler => "< 3" do
       simulate_ruby_engine "jruby" do
         gemfile <<-G
           source "file://#{gem_repo1}"
@@ -653,7 +671,7 @@ G
       end
     end
 
-    it "fails when patchlevel doesn't match" do
+    it "fails when patchlevel doesn't match", :bundler => "< 3" do
       gemfile <<-G
         source "file://#{gem_repo1}"
         gem "rack"
@@ -671,11 +689,10 @@ G
 
   context "bundle cache" do
     before do
-      gemfile <<-G
+      install_gemfile <<-G
+        source "file:#{gem_repo1}"
         gem 'rack'
       G
-
-      system_gems "rack-1.0.0"
     end
 
     it "copies the .gem file to vendor/cache when ruby version matches" do
@@ -691,13 +708,14 @@ G
 
     it "copies the .gem file to vendor/cache when ruby version matches for any engine" do
       simulate_ruby_engine "jruby" do
-        gemfile <<-G
+        install_gemfile! <<-G
+          source "file:#{gem_repo1}"
           gem 'rack'
 
           #{ruby_version_correct_engineless}
         G
 
-        bundle :cache
+        bundle! :cache
         expect(bundled_app("vendor/cache/rack-1.0.0.gem")).to exist
       end
     end
@@ -752,11 +770,10 @@ G
 
   context "bundle pack" do
     before do
-      gemfile <<-G
+      install_gemfile! <<-G
+        source "file:#{gem_repo1}"
         gem 'rack'
       G
-
-      system_gems "rack-1.0.0"
     end
 
     it "copies the .gem file to vendor/cache when ruby version matches" do
@@ -772,7 +789,8 @@ G
 
     it "copies the .gem file to vendor/cache when ruby version matches any engine" do
       simulate_ruby_engine "jruby" do
-        gemfile <<-G
+        install_gemfile! <<-G
+          source "file:#{gem_repo1}"
           gem 'rack'
 
           #{ruby_version_correct_engineless}
@@ -834,7 +852,7 @@ G
   context "bundle exec" do
     before do
       ENV["BUNDLER_FORCE_TTY"] = "true"
-      system_gems "rack-1.0.0", "rack-0.9.1"
+      system_gems "rack-1.0.0", "rack-0.9.1", :path => :bundle_path
     end
 
     it "activates the correct gem when ruby version matches" do
@@ -845,11 +863,12 @@ G
       G
 
       bundle "exec rackup"
-      expect(out).to eq("0.9.1")
+      expect(out).to include("0.9.1")
     end
 
     it "activates the correct gem when ruby version matches any engine" do
       simulate_ruby_engine "jruby" do
+        system_gems "rack-1.0.0", "rack-0.9.1", :path => :bundle_path
         gemfile <<-G
           gem "rack", "0.9.1"
 
@@ -857,7 +876,7 @@ G
         G
 
         bundle "exec rackup"
-        expect(out).to eq("0.9.1")
+        expect(out).to include("0.9.1")
       end
     end
 
@@ -909,7 +928,7 @@ G
     end
   end
 
-  context "bundle console" do
+  context "bundle console", :bundler => "< 3" do
     before do
       install_gemfile <<-G
         source "file://#{gem_repo1}"
@@ -1174,6 +1193,7 @@ G
 
     it "returns list of outdated gems when the ruby version matches for any engine" do
       simulate_ruby_engine "jruby" do
+        bundle! :install
         update_repo2 do
           build_gem "activesupport", "3.0"
           update_git "foo", :path => lib_path("foo")
