@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 require 'test/unit'
 require 'date'
 
@@ -180,6 +181,75 @@ class TestSH < Test::Unit::TestCase
 		 [d.year, d.mon, d.mday, d.hour, d.min, d.sec, d.offset])
   end
 
+  def test_fractional
+    d = Date.jd(2451944.0)
+    assert_equal(2451944, d.jd)
+    d = Date.jd(Rational(2451944))
+    assert_equal(2451944, d.jd)
+    d = Date.jd(2451944.5)
+    assert_equal(2451944, d.jd)
+    d = Date.jd(Rational('2451944.5'))
+    assert_equal(2451944, d.jd)
+
+    d = Date.civil(2001, 2, 3.0)
+    assert_equal([2001, 2, 3], [d.year, d.mon, d.mday])
+    d = Date.civil(2001, 2, Rational(3))
+    assert_equal([2001, 2, 3], [d.year, d.mon, d.mday])
+    d = Date.civil(2001, 2, 3.5)
+    assert_equal([2001, 2, 3], [d.year, d.mon, d.mday])
+    d = Date.civil(2001, 2, Rational('3.5'))
+    assert_equal([2001, 2, 3], [d.year, d.mon, d.mday])
+
+    d = Date.ordinal(2001, 2.0)
+    assert_equal([2001, 2], [d.year, d.yday])
+    d = Date.ordinal(2001, Rational(2))
+    assert_equal([2001, 2], [d.year, d.yday])
+
+    d = Date.commercial(2001, 2, 3.0)
+    assert_equal([2001, 2, 3], [d.cwyear, d.cweek, d.cwday])
+    d = Date.commercial(2001, 2, Rational(3))
+    assert_equal([2001, 2, 3], [d.cwyear, d.cweek, d.cwday])
+
+    d = DateTime.jd(2451944.0)
+    assert_equal(2451944, d.jd)
+    d = DateTime.jd(Rational(2451944))
+    assert_equal(2451944, d.jd)
+    d = DateTime.jd(2451944.5)
+    assert_equal([2451944, 12], [d.jd, d.hour])
+    d = DateTime.jd(Rational('2451944.5'))
+    assert_equal([2451944, 12], [d.jd, d.hour])
+
+    d = DateTime.civil(2001, 2, 3.0)
+    assert_equal([2001, 2, 3], [d.year, d.mon, d.mday])
+    d = DateTime.civil(2001, 2, Rational(3))
+    assert_equal([2001, 2, 3], [d.year, d.mon, d.mday])
+    d = DateTime.civil(2001, 2, 3.5)
+    assert_equal([2001, 2, 3, 12], [d.year, d.mon, d.mday, d.hour])
+    d = DateTime.civil(2001, 2, Rational('3.5'))
+    assert_equal([2001, 2, 3, 12], [d.year, d.mon, d.mday, d.hour])
+    d = DateTime.civil(2001, 2, 3, 4.5)
+    assert_equal([2001, 2, 3, 4, 30], [d.year, d.mon, d.mday, d.hour, d.min])
+    d = DateTime.civil(2001, 2, 3, Rational('4.5'))
+    assert_equal([2001, 2, 3, 4, 30], [d.year, d.mon, d.mday, d.hour, d.min])
+    d = DateTime.civil(2001, 2, 3, 4, 5.5)
+    assert_equal([2001, 2, 3, 4, 5, 30],
+		 [d.year, d.mon, d.mday, d.hour, d.min, d.sec])
+    d = DateTime.civil(2001, 2, 3, 4, Rational('5.5'))
+    assert_equal([2001, 2, 3, 4, 5, 30],
+		 [d.year, d.mon, d.mday, d.hour, d.min, d.sec])
+
+    d = DateTime.ordinal(2001, 2.0)
+    assert_equal([2001, 2], [d.year, d.yday])
+    d = DateTime.ordinal(2001, Rational(2))
+    assert_equal([2001, 2], [d.year, d.yday])
+
+    d = DateTime.commercial(2001, 2, 3.0)
+    assert_equal([2001, 2, 3], [d.cwyear, d.cweek, d.cwday])
+    d = DateTime.commercial(2001, 2, Rational(3))
+    assert_equal([2001, 2, 3], [d.cwyear, d.cweek, d.cwday])
+
+  end
+
   def test_canon24oc
     d = DateTime.jd(2451943,24)
     assert_equal([2001, 2, 3, 0, 0, 0, 0],
@@ -196,10 +266,8 @@ class TestSH < Test::Unit::TestCase
   end
 
   def test_zone
-    d = Date.new(2001, 2, 3)
-    assert_equal(Encoding::US_ASCII, d.send(:zone).encoding)
     d = DateTime.new(2001, 2, 3)
-    assert_equal(Encoding::US_ASCII, d.send(:zone).encoding)
+    assert_equal(Encoding::US_ASCII, d.zone.encoding)
   end
 
   def test_to_s
@@ -218,6 +286,17 @@ class TestSH < Test::Unit::TestCase
     assert_equal(Encoding::US_ASCII, d.inspect.encoding)
   end
 
+  def test_strftime
+    assert_raise(Errno::ERANGE) do
+      Date.today.strftime('%100000z')
+    end
+    assert_raise(Errno::ERANGE) do
+      Date.new(1 << 10000).strftime('%Y')
+    end
+    assert_equal('-3786825600', Date.new(1850).strftime('%s'))
+    assert_equal('-3786825600000', Date.new(1850).strftime('%Q'))
+  end
+
   def test_cmp
     assert_equal(-1, Date.new(2001,2,3) <=> Date.new(2001,2,4))
     assert_equal(0, Date.new(2001,2,3) <=> Date.new(2001,2,3))
@@ -232,6 +311,8 @@ class TestSH < Test::Unit::TestCase
     assert_equal(-1, Date.new(2001,2,3) <=> Rational('4903888/2'))
     assert_equal(0, Date.new(2001,2,3) <=> Rational('4903887/2'))
     assert_equal(1, Date.new(2001,2,3) <=> Rational('4903886/2'))
+
+    assert_equal(-1, Date.new(-4713,11,1,Date::GREGORIAN) <=> Date.new(-4713,12,1,Date::GREGORIAN))
   end
 
   def test_eqeqeq
@@ -394,11 +475,90 @@ class TestSH < Test::Unit::TestCase
     period2_iter(+cm_period * (1 << 64) - 3, +cm_period * (1 << 64) + 3)
   end
 
-  def test_marshal
+  def test_different_alignments
+    assert_equal(0, Date.jd(0) <=> Date.civil(-4713, 11, 24, Date::GREGORIAN))
+    assert_equal(0, Date.jd(213447717) <=> Date.civil(579687, 11, 24))
+    assert_equal(0, Date.jd(-213447717) <=> Date.civil(-589113, 11, 24, Date::GREGORIAN))
+
+    assert_equal(0, Date.jd(0) <=> DateTime.civil(-4713, 11, 24, 0, 0, 0, 0, Date::GREGORIAN))
+    assert_equal(0, Date.jd(213447717) <=> DateTime.civil(579687, 11, 24))
+    assert_equal(0, Date.jd(-213447717) <=> DateTime.civil(-589113, 11, 24, 0, 0, 0, 0, Date::GREGORIAN))
+
+    assert(Date.jd(0) == Date.civil(-4713, 11, 24, Date::GREGORIAN))
+    assert(Date.jd(213447717) == Date.civil(579687, 11, 24))
+    assert(Date.jd(-213447717) == Date.civil(-589113, 11, 24, Date::GREGORIAN))
+
+    assert(Date.jd(0) == DateTime.civil(-4713, 11, 24, 0, 0, 0, 0, Date::GREGORIAN))
+    assert(Date.jd(213447717) == DateTime.civil(579687, 11, 24))
+    assert(Date.jd(-213447717) == DateTime.civil(-589113, 11, 24, 0, 0, 0, 0, Date::GREGORIAN))
+
+    assert(Date.jd(0) === Date.civil(-4713, 11, 24, Date::GREGORIAN))
+    assert(Date.jd(213447717) === Date.civil(579687, 11, 24))
+    assert(Date.jd(-213447717) === Date.civil(-589113, 11, 24, Date::GREGORIAN))
+
+    assert(Date.jd(0) === DateTime.civil(-4713, 11, 24, 12, 0, 0, 0, Date::GREGORIAN))
+    assert(Date.jd(213447717) === DateTime.civil(579687, 11, 24, 12))
+    assert(Date.jd(-213447717) === DateTime.civil(-589113, 11, 24, 12, 0, 0, 0, Date::GREGORIAN))
+
+    a = Date.jd(0)
+    b = Date.civil(-4713, 11, 24, Date::GREGORIAN)
+    assert_equal(0, a <=> b)
+
+    a = Date.civil(-4712, 1, 1, Date::JULIAN)
+    b = Date.civil(-4713, 11, 24, Date::GREGORIAN)
+    a.jd; b.jd
+    assert_equal(0, a <=> b)
+
+    a = Date.jd(0)
+    b = Date.civil(-4713, 11, 24, Date::GREGORIAN)
+    assert(a == b)
+
+    a = Date.civil(-4712, 1, 1, Date::JULIAN)
+    b = Date.civil(-4713, 11, 24, Date::GREGORIAN)
+    a.jd; b.jd
+    assert(a == b)
+
+    a = Date.jd(0)
+    b = Date.civil(-4713, 11, 24, Date::GREGORIAN)
+    assert(a === b)
+
+    a = Date.civil(-4712, 1, 1, Date::JULIAN)
+    b = Date.civil(-4713, 11, 24, Date::GREGORIAN)
+    a.jd; b.jd
+    assert(a === b)
+  end
+
+  def test_marshal14
+    s = "\x04\x03u:\x01\x04Date\x01\v\x04\x03[\x01\x02i\x03\xE8i%T"
+    d = Marshal.load(s)
+    assert_equal(Rational(4903887,2), d.ajd)
+    assert_equal(Date::GREGORIAN, d.start)
+  end
+
+  def test_marshal16
+    s = "\x04\x06u:\tDate\x0F\x04\x06[\ai\x03\xE8i%T"
+    d = Marshal.load(s)
+    assert_equal(Rational(4903887,2), d.ajd)
+    assert_equal(Date::GREGORIAN, d.start)
+  end
+
+  def test_marshal18
+    s = "\x04\bu:\tDateP\x04\b[\bo:\rRational\a:\x0F@numeratori\x03\xCF\xD3J:\x11@denominatori\ai\x00o:\x13Date::Infinity\x06:\a@di\xFA"
+    d = Marshal.load(s)
+    assert_equal(Rational(4903887,2), d.ajd)
+    assert_equal(Date::GREGORIAN, d.start)
+
+    s = "\x04\bu:\rDateTime`\x04\b[\bo:\rRational\a:\x0F@numeratorl+\b\xC9\xB0\x81\xBD\x02\x00:\x11@denominatori\x02\xC0\x12o;\x00\a;\x06i\b;\ai\ro:\x13Date::Infinity\x06:\a@di\xFA"
+    d = Marshal.load(s)
+    assert_equal(Rational(11769327817,4800), d.ajd)
+    assert_equal(Rational(9,24), d.offset)
+    assert_equal(Date::GREGORIAN, d.start)
+  end
+
+  def test_marshal192
     s = "\x04\bU:\tDate[\bU:\rRational[\ai\x03\xCF\xD3Ji\ai\x00o:\x13Date::Infinity\x06:\a@di\xFA"
     d = Marshal.load(s)
     assert_equal(Rational(4903887,2), d.ajd)
-    assert_equal(Rational(0,24), d.send(:offset))
     assert_equal(Date::GREGORIAN, d.start)
 
     s = "\x04\bU:\rDateTime[\bU:\rRational[\al+\b\xC9\xB0\x81\xBD\x02\x00i\x02\xC0\x12U;\x06[\ai\bi\ro:\x13Date::Infinity\x06:\a@di\xFA"
@@ -411,29 +571,29 @@ class TestSH < Test::Unit::TestCase
   def test_taint
     h = Date._strptime('15:43+09:00', '%R%z')
     assert_equal(false, h[:zone].tainted?)
-    h = Date._strptime('15:43+09:00'.taint, '%R%z')
+    h = Date._strptime('15:43+09:00'.dup.taint, '%R%z')
     assert_equal(true, h[:zone].tainted?)
 
     h = Date._strptime('1;1/0', '%d')
     assert_equal(false, h[:leftover].tainted?)
-    h = Date._strptime('1;1/0'.taint, '%d')
+    h = Date._strptime('1;1/0'.dup.taint, '%d')
     assert_equal(true, h[:leftover].tainted?)
 
     h = Date._parse('15:43+09:00')
     assert_equal(false, h[:zone].tainted?)
-    h = Date._parse('15:43+09:00'.taint)
+    h = Date._parse('15:43+09:00'.dup.taint)
     assert_equal(true, h[:zone].tainted?)
 
     s = Date.today.strftime('new 105')
     assert_equal(false, s.tainted?)
-    s = Date.today.strftime('new 105'.taint)
+    s = Date.today.strftime('new 105'.dup.taint)
     assert_equal(true, s.tainted?)
-    s = Date.today.strftime("new \000 105".taint)
+    s = Date.today.strftime("new \000 105".dup.taint)
     assert_equal(true, s.tainted?)
 
     s = DateTime.now.strftime('super $record')
     assert_equal(false, s.tainted?)
-    s = DateTime.now.strftime('super $record'.taint)
+    s = DateTime.now.strftime('super $record'.dup.taint)
     assert_equal(true, s.tainted?)
   end
 
@@ -451,29 +611,29 @@ class TestSH < Test::Unit::TestCase
       assert_equal(Encoding::US_ASCII, s.encoding) if s
     end
 
-    h = Date._strptime('15:43+09:00'.force_encoding('euc-jp'), '%R%z')
+    h = Date._strptime('15:43+09:00'.dup.force_encoding('euc-jp'), '%R%z')
     assert_equal(Encoding::EUC_JP, h[:zone].encoding)
-    h = Date._strptime('15:43+09:00'.force_encoding('ascii-8bit'), '%R%z')
+    h = Date._strptime('15:43+09:00'.dup.force_encoding('ascii-8bit'), '%R%z')
     assert_equal(Encoding::ASCII_8BIT, h[:zone].encoding)
 
-    h = Date._strptime('1;1/0'.force_encoding('euc-jp'), '%d')
+    h = Date._strptime('1;1/0'.dup.force_encoding('euc-jp'), '%d')
     assert_equal(Encoding::EUC_JP, h[:leftover].encoding)
-    h = Date._strptime('1;1/0'.force_encoding('ascii-8bit'), '%d')
+    h = Date._strptime('1;1/0'.dup.force_encoding('ascii-8bit'), '%d')
     assert_equal(Encoding::ASCII_8BIT, h[:leftover].encoding)
 
-    h = Date._parse('15:43+09:00'.force_encoding('euc-jp'))
+    h = Date._parse('15:43+09:00'.dup.force_encoding('euc-jp'))
     assert_equal(Encoding::EUC_JP, h[:zone].encoding)
-    h = Date._parse('15:43+09:00'.force_encoding('ascii-8bit'))
+    h = Date._parse('15:43+09:00'.dup.force_encoding('ascii-8bit'))
     assert_equal(Encoding::ASCII_8BIT, h[:zone].encoding)
 
-    s = Date.today.strftime('new 105'.force_encoding('euc-jp'))
+    s = Date.today.strftime('new 105'.dup.force_encoding('euc-jp'))
     assert_equal(Encoding::EUC_JP, s.encoding)
-    s = Date.today.strftime('new 105'.force_encoding('ascii-8bit'))
+    s = Date.today.strftime('new 105'.dup.force_encoding('ascii-8bit'))
     assert_equal(Encoding::ASCII_8BIT, s.encoding)
 
-    s = DateTime.now.strftime('super $record'.force_encoding('euc-jp'))
+    s = DateTime.now.strftime('super $record'.dup.force_encoding('euc-jp'))
     assert_equal(Encoding::EUC_JP, s.encoding)
-    s = DateTime.now.strftime('super $record'.force_encoding('ascii-8bit'))
+    s = DateTime.now.strftime('super $record'.dup.force_encoding('ascii-8bit'))
     assert_equal(Encoding::ASCII_8BIT, s.encoding)
   end
 
@@ -492,8 +652,7 @@ class TestSH < Test::Unit::TestCase
   end
 
   def test_base
-    skip unless defined?(Date.test_all)
     assert_equal(true, Date.test_all)
-  end
+  end if defined?(Date.test_all)
 
 end
