@@ -89,11 +89,13 @@ module Test
           }
 
           return e
+        ensure
+          unless e
+            exp = exp.first if exp.size == 1
+
+            flunk(message(msg) {"#{mu_pp(exp)} expected but nothing was raised"})
+          end
         end
-
-        exp = exp.first if exp.size == 1
-
-        flunk(message(msg) {"#{mu_pp(exp)} expected but nothing was raised"})
       end
 
       def assert_raises(*exp, &b)
@@ -472,7 +474,7 @@ EOT
         failed = []
         obj.each do |*a, &b|
           if blk.call(*a, &b)
-            failed << a.size > 1 ? a : a[0]
+            failed << (a.size > 1 ? a : a[0])
           end
         end
         assert(failed.empty?, message(m) {failed.pretty_inspect})
@@ -498,7 +500,9 @@ EOT
         end
       end
 
-      def prepare_syntax_check(code, fname = caller_locations(2, 1)[0], mesg = fname.to_s, verbose: nil)
+      def prepare_syntax_check(code, fname = nil, mesg = nil, verbose: nil)
+        fname ||= caller_locations(2, 1)[0]
+        mesg ||= fname.to_s
         verbose, $VERBOSE = $VERBOSE, verbose
         case
         when Array === fname
@@ -508,13 +512,19 @@ EOT
         else
           line = 1
         end
-        yield(code, fname, line, mesg)
+        yield(code, fname, line, message(mesg) {
+                if code.end_with?("\n")
+                  "```\n#{code}```\n"
+                else
+                  "```\n#{code}\n```\n""no-newline"
+                end
+              })
       ensure
         $VERBOSE = verbose
       end
 
-      def assert_valid_syntax(code, *args)
-        prepare_syntax_check(code, *args) do |src, fname, line, mesg|
+      def assert_valid_syntax(code, *args, **opt)
+        prepare_syntax_check(code, *args, **opt) do |src, fname, line, mesg|
           yield if defined?(yield)
           assert_nothing_raised(SyntaxError, mesg) do
             assert_equal(:ok, syntax_check(src, fname, line), mesg)
@@ -581,6 +591,8 @@ EOT
 
       def assert_in_out_err(args, test_stdin = "", test_stdout = [], test_stderr = [], message = nil,
                             success: nil, **opt)
+        args = Array(args).dup
+        args.insert((Hash === args[0] ? 1 : 0), '--disable=gems')
         stdout, stderr, status = EnvUtil.invoke_ruby(args, test_stdin, true, true, **opt)
         if signo = status.termsig
           EnvUtil.diagnostic_reports(Signal.signame(signo), status.pid, Time.now)
@@ -676,8 +688,8 @@ eom
       end
 
       def assert_warning(pat, msg = nil)
-        stderr = EnvUtil.verbose_warning {
-          EnvUtil.with_default_internal(pat.encoding) {
+        stderr = EnvUtil.with_default_internal(pat.encoding) {
+          EnvUtil.verbose_warning {
             yield
           }
         }
@@ -689,7 +701,20 @@ eom
         assert_warning(*args) {$VERBOSE = false; yield}
       end
 
+      def assert_no_warning(pat, msg = nil)
+        stderr = EnvUtil.verbose_warning {
+          EnvUtil.with_default_internal(pat.encoding) {
+            yield
+          }
+        }
+        msg = message(msg) {diff pat, stderr}
+        refute(pat === stderr, msg)
+      end
+
       def assert_no_memory_leak(args, prepare, code, message=nil, limit: 2.0, rss: false, **opt)
+        # TODO: consider choosing some appropriate limit for MJIT and stop skipping this once it does not randomly fail
+        skip 'assert_no_memory_leak may consider MJIT memory usage as leak' if defined?(RubyVM::MJIT) && RubyVM::MJIT.enabled?
+
         require_relative '../../memory_status'
         raise MiniTest::Skip, "unsupported platform" unless defined?(Memory::Status)
 
@@ -744,7 +769,7 @@ eom
         # [ruby-core:81540]
         min_hz = windows? ? 67 : 100
         min_measurable = 1.0 / min_hz
-        min_measurable *= 1.10 # add a little (10%) to account for misc. overheads
+        min_measurable *= 1.30 # add a little (30%) to account for misc. overheads
         if max < min_measurable
           max = min_measurable
         end
