@@ -1,14 +1,10 @@
 # frozen_string_literal: false
-begin
-  require "readline"
-rescue LoadError
-else
-  require "test/unit"
-  require "tempfile"
-  require "timeout"
-end
+require_relative "helper"
+require "test/unit"
+require "tempfile"
+require "timeout"
 
-class TestReadline < Test::Unit::TestCase
+module BasetestReadline
   INPUTRC = "INPUTRC"
   SAVED_ENV = %w[COLUMNS LINES]
 
@@ -30,89 +26,90 @@ class TestReadline < Test::Unit::TestCase
     SAVED_ENV.each_with_index {|k, i| ENV[k] = @saved_env[i] }
   end
 
-  if !/EditLine/n.match(Readline::VERSION)
-    def test_readline
-      with_temp_stdio do |stdin, stdout|
-        stdin.write("hello\n")
-        stdin.close
-        stdout.flush
-        line = replace_stdio(stdin.path, stdout.path) {
-          Readline.readline("> ", true)
-        }
-        assert_equal("hello", line)
-        assert_equal(true, line.tainted?)
-        stdout.rewind
-        assert_equal("> ", stdout.read(2))
-        assert_equal(1, Readline::HISTORY.length)
-        assert_equal("hello", Readline::HISTORY[0])
-        Thread.start {
-          $SAFE = 1
-          assert_raise(SecurityError) do
-            replace_stdio(stdin.path, stdout.path) do
-              Readline.readline("> ".taint)
-            end
+  def test_readline
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
+    with_temp_stdio do |stdin, stdout|
+      stdin.write("hello\n")
+      stdin.close
+      stdout.flush
+      line = replace_stdio(stdin.path, stdout.path) {
+        Readline.readline("> ", true)
+      }
+      assert_equal("hello", line)
+      assert_equal(true, line.tainted?)
+      stdout.rewind
+      assert_equal("> ", stdout.read(2))
+      assert_equal(1, Readline::HISTORY.length)
+      assert_equal("hello", Readline::HISTORY[0])
+      Thread.start {
+        $SAFE = 1
+        assert_raise(SecurityError) do
+          replace_stdio(stdin.path, stdout.path) do
+            Readline.readline("> ".taint)
           end
-        }.join
-      ensure
-        $SAFE = 0
-      end
+        end
+      }.join
+    ensure
+      $SAFE = 0
+    end
+  end
+
+  # line_buffer
+  # point
+  def test_line_buffer__point
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
+    skip "GNU Readline has special behaviors" if defined?(Reline) and Readline == Reline
+    begin
+      Readline.line_buffer
+      Readline.point
+    rescue NotImplementedError
+      return
     end
 
-    # line_buffer
-    # point
-    def test_line_buffer__point
-      begin
-        Readline.line_buffer
-        Readline.point
-      rescue NotImplementedError
-        return
-      end
-
-      with_temp_stdio do |stdin, stdout|
-        actual_text = nil
-        actual_line_buffer = nil
-        actual_point = nil
-        Readline.completion_proc = ->(text) {
-          actual_text = text
-          actual_point = Readline.point
-          actual_line_buffer = Readline.line_buffer
-          stdin.write(" finish\n")
-          stdin.flush
-          stdout.flush
-          return ["complete"]
-        }
-
-        stdin.write("first second\t")
+    with_temp_stdio do |stdin, stdout|
+      actual_text = nil
+      actual_line_buffer = nil
+      actual_point = nil
+      Readline.completion_proc = ->(text) {
+        actual_text = text
+        actual_point = Readline.point
+        actual_line_buffer = Readline.line_buffer
+        stdin.write(" finish\n")
         stdin.flush
-        Readline.completion_append_character = " "
-        replace_stdio(stdin.path, stdout.path) {
-          Readline.readline("> ", false)
-        }
-        assert_equal("second", actual_text)
-        assert_equal("first second", actual_line_buffer)
-        assert_equal(12, actual_point)
-        assert_equal("first complete  finish", Readline.line_buffer)
-        assert_equal(Encoding.find("locale"), Readline.line_buffer.encoding)
-        assert_equal(true, Readline.line_buffer.tainted?)
-        assert_equal(22, Readline.point)
+        stdout.flush
+        return ["complete"]
+      }
 
-        stdin.rewind
-        stdout.rewind
+      stdin.write("first second\t")
+      stdin.flush
+      Readline.completion_append_character = " "
+      replace_stdio(stdin.path, stdout.path) {
+        Readline.readline("> ", false)
+      }
+      assert_equal("second", actual_text)
+      assert_equal("first second", actual_line_buffer)
+      assert_equal(12, actual_point)
+      assert_equal("first complete  finish", Readline.line_buffer)
+      assert_equal(Encoding.find("locale"), Readline.line_buffer.encoding)
+      assert_equal(true, Readline.line_buffer.tainted?)
+      assert_equal(22, Readline.point)
 
-        stdin.write("first second\t")
-        stdin.flush
-        Readline.completion_append_character = nil
-        replace_stdio(stdin.path, stdout.path) {
-          Readline.readline("> ", false)
-        }
-        assert_equal("second", actual_text)
-        assert_equal("first second", actual_line_buffer)
-        assert_equal(12, actual_point)
-        assert_equal("first complete finish", Readline.line_buffer)
-        assert_equal(Encoding.find("locale"), Readline.line_buffer.encoding)
-        assert_equal(true, Readline.line_buffer.tainted?)
-        assert_equal(21, Readline.point)
-      end
+      stdin.rewind
+      stdout.rewind
+
+      stdin.write("first second\t")
+      stdin.flush
+      Readline.completion_append_character = nil
+      replace_stdio(stdin.path, stdout.path) {
+        Readline.readline("> ", false)
+      }
+      assert_equal("second", actual_text)
+      assert_equal("first second", actual_line_buffer)
+      assert_equal(12, actual_point)
+      assert_equal("first complete finish", Readline.line_buffer)
+      assert_equal(Encoding.find("locale"), Readline.line_buffer.encoding)
+      assert_equal(true, Readline.line_buffer.tainted?)
+      assert_equal(21, Readline.point)
     end
   end
 
@@ -147,6 +144,7 @@ class TestReadline < Test::Unit::TestCase
   end
 
   def test_completion_proc_empty_result
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
     with_temp_stdio do |stdin, stdout|
       stdin.write("first\t")
       stdin.flush
@@ -165,7 +163,7 @@ class TestReadline < Test::Unit::TestCase
       rescue NotimplementedError
       end
     end
-  end if !/EditLine/n.match(Readline::VERSION)
+  end
 
   def test_get_screen_size
     begin
@@ -225,6 +223,7 @@ class TestReadline < Test::Unit::TestCase
   end
 
   def test_completion_encoding
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
     bug5941 = '[Bug #5941]'
     append_character = Readline.completion_append_character
     Readline.completion_append_character = ""
@@ -264,9 +263,10 @@ class TestReadline < Test::Unit::TestCase
       with_pipe {|r, w| w << "\t"}
     end
   ensure
+    return if /EditLine/n.match(Readline::VERSION)
     Readline.completion_case_fold = completion_case_fold
     Readline.completion_append_character = append_character
-  end if !/EditLine/n.match(Readline::VERSION)
+  end
 
   # basic_word_break_characters
   # completer_word_break_characters
@@ -293,7 +293,7 @@ class TestReadline < Test::Unit::TestCase
             Readline.send((method_name + "=").to_sym, e)
             res = Readline.send(method_name.to_sym)
             assert_equal(e, res)
-            assert_equal(enc, res.encoding)
+            assert_equal(enc, res.encoding, "Readline.#{method_name} should be #{enc.name}")
           end
         ensure
           Readline.send((method_name + "=").to_sym, saved) if saved
@@ -325,6 +325,7 @@ class TestReadline < Test::Unit::TestCase
   end
 
   def test_point
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
     assert_equal(0, Readline.point)
     Readline.insert_text('12345')
     assert_equal(5, Readline.point)
@@ -336,9 +337,10 @@ class TestReadline < Test::Unit::TestCase
 
     assert_equal('1234abc5', Readline.line_buffer)
   rescue NotImplementedError
-  end if !/EditLine/n.match(Readline::VERSION)
+  end
 
   def test_insert_text
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
     str = "test_insert_text"
     assert_equal(0, Readline.point)
     assert_equal(Readline, Readline.insert_text(str))
@@ -366,9 +368,10 @@ class TestReadline < Test::Unit::TestCase
     Readline.delete_text
     assert_equal("", Readline.line_buffer)
   rescue NotImplementedError
-  end if !/EditLine/n.match(Readline::VERSION)
+  end
 
   def test_delete_text
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
     str = "test_insert_text"
     assert_equal(0, Readline.point)
     assert_equal(Readline, Readline.insert_text(str))
@@ -376,16 +379,19 @@ class TestReadline < Test::Unit::TestCase
     assert_equal(str, Readline.line_buffer)
     Readline.delete_text
 
-    # NOTE: unexpected but GNU Readline's spec
-    assert_equal(16, Readline.point)
-    assert_equal("", Readline.line_buffer)
-    assert_equal(Readline, Readline.insert_text(str))
-    assert_equal(32, Readline.point)
-    assert_equal("", Readline.line_buffer)
+    if !defined?(Reline) or Readline != Reline
+      # NOTE: unexpected but GNU Readline's spec
+      assert_equal(16, Readline.point)
+      assert_equal("", Readline.line_buffer)
+      assert_equal(Readline, Readline.insert_text(str))
+      assert_equal(32, Readline.point)
+      assert_equal("", Readline.line_buffer)
+    end
   rescue NotImplementedError
-  end if !/EditLine/n.match(Readline::VERSION)
+  end
 
   def test_modify_text_in_pre_input_hook
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
     with_temp_stdio {|stdin, stdout|
       begin
         stdin.write("world\n")
@@ -399,7 +405,11 @@ class TestReadline < Test::Unit::TestCase
           line = Readline.readline("> ")
           assert_equal("hello world", line)
         end
-        assert_equal("> hello world\n", stdout.read)
+        # Readline 4.3 doesn't include inserted text or input
+        # Reline's rendering logic is tricky
+        if Readline::VERSION != '4.3' and (!defined?(Reline) or Readline != Reline)
+          assert_equal("> hello world\n", stdout.read)
+        end
         stdout.close
       rescue NotImplementedError
       ensure
@@ -409,9 +419,11 @@ class TestReadline < Test::Unit::TestCase
         end
       end
     }
-  end if !/EditLine|\A4\.3\z/n.match(Readline::VERSION)
+  end
 
   def test_input_metachar
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
+    skip("Won't pass on mingw w/readline 7.0.005 [ruby-core:45682]") if mingw?
     bug6601 = '[ruby-core:45682]'
     Readline::HISTORY << "hello"
     wo = nil
@@ -421,12 +433,14 @@ class TestReadline < Test::Unit::TestCase
     end
     assert_equal("hello", line, bug6601)
   ensure
-    wo.close
+    wo&.close
+    return if /EditLine/n.match(Readline::VERSION)
     Readline.delete_text
     Readline::HISTORY.clear
-  end if !/EditLine/n.match(Readline::VERSION)
+  end
 
   def test_input_metachar_multibyte
+    skip "Skip Editline" if /EditLine/n.match(Readline::VERSION)
     unless Encoding.find("locale") == Encoding::UTF_8
       return if assert_under_utf8
       skip 'this test needs UTF-8 locale'
@@ -450,11 +464,13 @@ class TestReadline < Test::Unit::TestCase
       end
     end
   ensure
+    return if /EditLine/n.match(Readline::VERSION)
     Readline.delete_text
     Readline::HISTORY.clear
-  end if !/EditLine/n.match(Readline::VERSION)
+  end
 
   def test_refresh_line
+    skip "Only when refresh_line exists" unless Readline.respond_to?(:refresh_line)
     bug6232 = '[ruby-core:43957] [Bug #6232] refresh_line after set_screen_size'
     with_temp_stdio do |stdin, stdout|
       replace_stdio(stdin.path, stdout.path) do
@@ -464,7 +480,7 @@ class TestReadline < Test::Unit::TestCase
         end;
       end
     end
-  end if Readline.respond_to?(:refresh_line)
+  end
 
   def test_setting_quoting_detection_proc
     return unless Readline.respond_to?(:quoting_detection_proc=)
@@ -552,6 +568,69 @@ class TestReadline < Test::Unit::TestCase
     Readline.completer_word_break_characters = saved_completer_word_break_characters
   end
 
+  def test_completion_quote_character_completing_unquoted_argument
+    return unless Readline.respond_to?(:completion_quote_character)
+
+    quote_character = "original value"
+    Readline.completion_proc = -> (_) do
+      quote_character = Readline.completion_quote_character
+      []
+    end
+    Readline.completer_quote_characters = "'\""
+
+    with_temp_stdio do |stdin, stdout|
+      replace_stdio(stdin.path, stdout.path) do
+        stdin.write("input\t")
+        stdin.flush
+        Readline.readline("> ", false)
+      end
+    end
+
+    assert_nil(quote_character)
+  end
+
+  def test_completion_quote_character_completing_quoted_argument
+    return unless Readline.respond_to?(:completion_quote_character)
+
+    quote_character = "original value"
+    Readline.completion_proc = -> (_) do
+      quote_character = Readline.completion_quote_character
+      []
+    end
+    Readline.completer_quote_characters = "'\""
+
+    with_temp_stdio do |stdin, stdout|
+      replace_stdio(stdin.path, stdout.path) do
+        stdin.write("'input\t")
+        stdin.flush
+        Readline.readline("> ", false)
+      end
+    end
+
+    assert_equal("'", quote_character)
+  end
+
+  def test_completion_quote_character_after_completion
+    return unless Readline.respond_to?(:completion_quote_character)
+    if /solaris/i =~ RUBY_PLATFORM
+      # http://rubyci.s3.amazonaws.com/solaris11s-sunc/ruby-trunk/log/20181228T102505Z.fail.html.gz
+      skip 'This test does not succeed on Oracle Developer Studio for now'
+    end
+
+    Readline.completion_proc = -> (_) { [] }
+    Readline.completer_quote_characters = "'\""
+
+    with_temp_stdio do |stdin, stdout|
+      replace_stdio(stdin.path, stdout.path) do
+        stdin.write("'input\t")
+        stdin.flush
+        Readline.readline("> ", false)
+      end
+    end
+
+    assert_nil(Readline.completion_quote_character)
+  end
+
   private
 
   def replace_stdio(stdin_path, stdout_path)
@@ -583,6 +662,11 @@ class TestReadline < Test::Unit::TestCase
     Tempfile.create("test_readline_stdin") {|stdin|
       Tempfile.create("test_readline_stdout") {|stdout|
         yield stdin, stdout
+        if windows?
+          # needed since readline holds refs to tempfiles, can't delete on Windows
+          Readline.input = STDIN
+          Readline.output = STDOUT
+        end
       }
     }
   end
@@ -619,4 +703,22 @@ class TestReadline < Test::Unit::TestCase
 SRC
     return true
   end
-end if defined?(::Readline)
+end
+
+class TestReadline < Test::Unit::TestCase
+  include BasetestReadline
+
+  def setup
+    use_ext_readline
+    super
+  end
+end if defined?(ReadlineSo)
+
+class TestRelineAsReadline < Test::Unit::TestCase
+  include BasetestReadline
+
+  def setup
+    use_lib_reline
+    super
+  end
+end
