@@ -2,7 +2,7 @@
 
   version.c -
 
-  $Author: nobu $
+  $Author$
   created at: Thu Sep 30 20:08:01 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -11,6 +11,8 @@
 
 #include "ruby/ruby.h"
 #include "version.h"
+#include "vm_core.h"
+#include "mjit.h"
 #include <stdio.h>
 
 #ifndef EXIT_SUCCESS
@@ -26,22 +28,30 @@ const int ruby_api_version[] = {
     RUBY_API_VERSION_MINOR,
     RUBY_API_VERSION_TEENY,
 };
+#define RUBY_VERSION \
+    STRINGIZE(RUBY_VERSION_MAJOR) "." \
+    STRINGIZE(RUBY_VERSION_MINOR) "." \
+    STRINGIZE(RUBY_VERSION_TEENY) ""
+#ifndef RUBY_FULL_REVISION
+# define RUBY_FULL_REVISION RUBY_REVISION
+#endif
 const char ruby_version[] = RUBY_VERSION;
+const char ruby_revision[] = RUBY_FULL_REVISION;
 const char ruby_release_date[] = RUBY_RELEASE_DATE;
 const char ruby_platform[] = RUBY_PLATFORM;
 const int ruby_patchlevel = RUBY_PATCHLEVEL;
-const char ruby_description[] = RUBY_DESCRIPTION;
+const char ruby_description[] = RUBY_DESCRIPTION_WITH("");
+static const char ruby_description_with_jit[] = RUBY_DESCRIPTION_WITH(" +JIT");
 const char ruby_copyright[] = RUBY_COPYRIGHT;
 const char ruby_engine[] = "ruby";
-VALUE ruby_engine_name = Qnil;
 
 /*! Defines platform-depended Ruby-level constants */
 void
 Init_version(void)
 {
     enum {ruby_patchlevel = RUBY_PATCHLEVEL};
-    enum {ruby_revision = RUBY_REVISION};
     VALUE version;
+    VALUE ruby_engine_name;
     /*
      * The running version of ruby
      */
@@ -62,11 +72,7 @@ Init_version(void)
     /*
      * The SVN revision for this ruby.
      */
-    rb_define_global_const("RUBY_REVISION", MKINT(revision));
-    /*
-     * The full ruby version string, like <tt>ruby -v</tt> prints'
-     */
-    rb_define_global_const("RUBY_DESCRIPTION", MKSTR(description));
+    rb_define_global_const("RUBY_REVISION", MKSTR(revision));
     /*
      * The copyright string for ruby
      */
@@ -75,17 +81,47 @@ Init_version(void)
      * The engine or interpreter this ruby uses.
      */
     rb_define_global_const("RUBY_ENGINE", ruby_engine_name = MKSTR(engine));
+    ruby_set_script_name(ruby_engine_name);
     /*
      * The version of the engine or interpreter this ruby uses.
      */
     rb_define_global_const("RUBY_ENGINE_VERSION", (1 ? version : MKSTR(version)));
 }
 
+#if USE_MJIT
+#define MJIT_OPTS_ON mjit_opts.on
+#else
+#define MJIT_OPTS_ON 0
+#endif
+
+void
+Init_ruby_description(void)
+{
+    VALUE description;
+
+    if (MJIT_OPTS_ON) {
+        description = MKSTR(description_with_jit);
+    }
+    else {
+        description = MKSTR(description);
+    }
+
+    /*
+     * The full ruby version string, like <tt>ruby -v</tt> prints
+     */
+    rb_define_global_const("RUBY_DESCRIPTION", /* MKSTR(description) */ description);
+}
+
 /*! Prints the version information of the CRuby interpreter to stdout. */
 void
 ruby_show_version(void)
 {
-    PRINT(description);
+    if (MJIT_OPTS_ON) {
+        PRINT(description_with_jit);
+    }
+    else {
+        PRINT(description);
+    }
 #ifdef RUBY_LAST_COMMIT_TITLE
     fputs("last_commit=" RUBY_LAST_COMMIT_TITLE, stdout);
 #endif
@@ -95,12 +131,10 @@ ruby_show_version(void)
     fflush(stdout);
 }
 
-/*! Prints the copyright notice of the CRuby interpreter to stdout and \em exits
- *  this process successfully.
- */
+/*! Prints the copyright notice of the CRuby interpreter to stdout. */
 void
 ruby_show_copyright(void)
 {
     PRINT(copyright);
-    exit(EXIT_SUCCESS);
+    fflush(stdout);
 }
