@@ -19,10 +19,11 @@ struct args_info {
     /* basic args info */
     VALUE *argv;
     int argc;
-    const struct rb_call_info_kw_arg *kw_arg;
 
     /* additional args info */
     int rest_index;
+    int rest_dupped;
+    const struct rb_call_info_kw_arg *kw_arg;
     VALUE *kw_argv;
     VALUE rest;
 };
@@ -32,6 +33,15 @@ enum arg_setup_type {
     arg_setup_block
 };
 
+static inline void
+arg_rest_dup(struct args_info *args)
+{
+    if (!args->rest_dupped) {
+        args->rest = rb_ary_dup(args->rest);
+        args->rest_dupped = TRUE;
+    }
+}
+
 static inline int
 args_argc(struct args_info *args)
 {
@@ -49,7 +59,7 @@ args_extend(struct args_info *args, const int min_argc)
     int i;
 
     if (args->rest) {
-	args->rest = rb_ary_dup(args->rest);
+        arg_rest_dup(args);
 	VM_ASSERT(args->rest_index == 0);
 	for (i=args->argc + RARRAY_LENINT(args->rest); i<min_argc; i++) {
 	    rb_ary_push(args->rest, Qnil);
@@ -69,7 +79,7 @@ args_reduce(struct args_info *args, int over_argc)
 	const long len = RARRAY_LEN(args->rest);
 
 	if (len > over_argc) {
-	    args->rest = rb_ary_dup(args->rest);
+	    arg_rest_dup(args);
 	    rb_ary_resize(args->rest, len - over_argc);
 	    return;
 	}
@@ -114,7 +124,7 @@ args_copy(struct args_info *args)
     if (args->rest != Qfalse) {
 	int argc = args->argc;
 	args->argc = 0;
-	args->rest = rb_ary_dup(args->rest); /* make dup */
+        arg_rest_dup(args);
 
 	/*
 	 * argv: [m0, m1, m2, m3]
@@ -146,6 +156,7 @@ args_copy(struct args_info *args)
     else if (args->argc > 0) {
 	args->rest = rb_ary_new_from_values(args->argc, args->argv);
 	args->rest_index = 0;
+        args->rest_dupped = TRUE;
 	args->argc = 0;
     }
 }
@@ -153,7 +164,7 @@ args_copy(struct args_info *args)
 static inline const VALUE *
 args_rest_argv(struct args_info *args)
 {
-    return RARRAY_CONST_PTR(args->rest) + args->rest_index;
+    return RARRAY_CONST_PTR_TRANSIENT(args->rest) + args->rest_index;
 }
 
 static inline VALUE
@@ -162,7 +173,8 @@ args_rest_array(struct args_info *args)
     VALUE ary;
 
     if (args->rest) {
-	ary = rb_ary_subseq(args->rest, args->rest_index, RARRAY_LEN(args->rest) - args->rest_index);
+        ary = rb_ary_behead(args->rest, args->rest_index);
+        args->rest_index = 0;
 	args->rest = 0;
     }
     else {
@@ -219,7 +231,7 @@ args_pop_keyword_hash(struct args_info *args, VALUE *kw_hash_ptr)
 		    RARRAY_ASET(args->rest, len - 1, rest_hash);
 		}
 		else {
-		    args->rest = rb_ary_dup(args->rest);
+		    arg_rest_dup(args);
 		    rb_ary_pop(args->rest);
 		    return TRUE;
 		}
@@ -269,7 +281,7 @@ args_stored_kw_argv_to_hash(struct args_info *args)
     args->kw_argv = NULL;
 
     if (args->rest) {
-	args->rest = rb_ary_dup(args->rest);
+	arg_rest_dup(args);
 	rb_ary_push(args->rest, h);
     }
     else {
@@ -301,9 +313,8 @@ static inline void
 args_setup_post_parameters(struct args_info *args, int argc, VALUE *locals)
 {
     long len;
-    args_copy(args);
     len = RARRAY_LEN(args->rest);
-    MEMCPY(locals, RARRAY_CONST_PTR(args->rest) + len - argc, VALUE, argc);
+    MEMCPY(locals, RARRAY_CONST_PTR_TRANSIENT(args->rest) + len - argc, VALUE, argc);
     rb_ary_resize(args->rest, len - argc);
 }
 
@@ -323,13 +334,13 @@ args_setup_opt_parameters(struct args_info *args, int opt_max, VALUE *locals)
 	args->argc = 0;
 
 	if (args->rest) {
-	    int len = RARRAY_LENINT(args->rest);
-	    const VALUE *argv = RARRAY_CONST_PTR(args->rest);
+            int len = RARRAY_LENINT(args->rest);
+            const VALUE *argv = RARRAY_CONST_PTR_TRANSIENT(args->rest);
 
-	    for (; i<opt_max && args->rest_index < len; i++, args->rest_index++) {
-		locals[i] = argv[args->rest_index];
-	    }
-	}
+            for (; i<opt_max && args->rest_index < len; i++, args->rest_index++) {
+                locals[i] = argv[args->rest_index];
+            }
+        }
 
 	/* initialize by nil */
 	for (j=i; j<opt_max; j++) {
@@ -343,7 +354,6 @@ args_setup_opt_parameters(struct args_info *args, int opt_max, VALUE *locals)
 static inline void
 args_setup_rest_parameter(struct args_info *args, VALUE *locals)
 {
-    args_copy(args);
     *locals = args_rest_array(args);
 }
 
@@ -509,11 +519,13 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
     const int max_argc = (iseq->body->param.flags.has_rest == FALSE) ? min_argc + iseq->body->param.opt_num : UNLIMITED_ARGUMENTS;
     int opt_pc = 0;
     int given_argc;
+    int kw_splat = FALSE;
     struct args_info args_body, *args;
     VALUE keyword_hash = Qnil;
     VALUE * const orig_sp = ec->cfp->sp;
     unsigned int i;
 
+    vm_check_canary(ec, orig_sp);
     /*
      * Extend SP for GC.
      *
@@ -537,6 +549,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
     args = &args_body;
     given_argc = args->argc = calling->argc;
     args->argv = locals;
+    args->rest_dupped = FALSE;
 
     if (ci->flag & VM_CALL_KWARG) {
 	args->kw_arg = ((struct rb_call_info_with_kwarg *)ci)->kw_arg;
@@ -600,10 +613,12 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 	}
     }
 
+    if (ci->flag & VM_CALL_KW_SPLAT) {
+	kw_splat = !iseq->body->param.flags.has_rest;
+    }
     if (given_argc > min_argc &&
 	(iseq->body->param.flags.has_kw || iseq->body->param.flags.has_kwrest ||
-	 (!iseq->body->param.flags.has_rest && given_argc > max_argc &&
-	  (ci->flag & VM_CALL_KW_SPLAT))) &&
+	 (kw_splat && given_argc > max_argc)) &&
 	args->kw_argv == NULL) {
 	if (args_pop_keyword_hash(args, &keyword_hash)) {
 	    given_argc--;
@@ -625,6 +640,10 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 	args_setup_lead_parameters(args, iseq->body->param.lead_num, locals + 0);
     }
 
+    if (iseq->body->param.flags.has_rest || iseq->body->param.flags.has_post){
+        args_copy(args);
+    }
+
     if (iseq->body->param.flags.has_post) {
 	args_setup_post_parameters(args, iseq->body->param.post_num, locals + iseq->body->param.post_start);
     }
@@ -667,6 +686,17 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
     else if (!NIL_P(keyword_hash) && RHASH_SIZE(keyword_hash) > 0) {
 	argument_kw_error(ec, iseq, "unknown", rb_hash_keys(keyword_hash));
     }
+    else if (kw_splat && NIL_P(keyword_hash)) {
+	if (RTEST(ruby_verbose)) {
+	    VALUE path = rb_iseq_path(iseq);
+	    VALUE line = rb_iseq_first_lineno(iseq);
+	    VALUE label = rb_iseq_label(iseq);
+	    rb_compile_warning(NIL_P(path) ? NULL : RSTRING_PTR(path), FIX2INT(line),
+			       "in `%s': the last argument was passed as a single Hash",
+			       NIL_P(label) ? NULL : RSTRING_PTR(label));
+	    rb_warning("although a splat keyword arguments here");
+	}
+    }
 
     if (iseq->body->param.flags.has_block) {
 	if (iseq->body->local_iseq == iseq) {
@@ -753,18 +783,19 @@ vm_caller_setup_arg_splat(rb_control_frame_t *cfp, struct rb_calling_info *calli
     VALUE *argv = cfp->sp - argc;
     VALUE ary = argv[argc-1];
 
+    vm_check_canary(GET_EC(), cfp->sp);
     cfp->sp--;
 
     if (!NIL_P(ary)) {
-	const VALUE *ptr = RARRAY_CONST_PTR(ary);
-	long len = RARRAY_LEN(ary), i;
+        const VALUE *ptr = RARRAY_CONST_PTR_TRANSIENT(ary);
+        long len = RARRAY_LEN(ary), i;
 
-	CHECK_VM_STACK_OVERFLOW(cfp, len);
+        CHECK_VM_STACK_OVERFLOW(cfp, len);
 
-	for (i = 0; i < len; i++) {
-	    *cfp->sp++ = ptr[i];
-	}
-	calling->argc += i - 1;
+        for (i = 0; i < len; i++) {
+            *cfp->sp++ = ptr[i];
+        }
+        calling->argc += i - 1;
     }
 }
 
@@ -796,7 +827,7 @@ vm_to_proc(VALUE proc)
 	    rb_callable_method_entry_with_refinements(CLASS_OF(proc), idTo_proc, NULL);
 
 	if (me) {
-	    b = vm_call0(GET_EC(), proc, idTo_proc, 0, NULL, me);
+            b = rb_vm_call0(GET_EC(), proc, idTo_proc, 0, NULL, me);
 	}
 	else {
 	    /* NOTE: calling method_missing */
@@ -822,13 +853,24 @@ refine_sym_proc_call(RB_BLOCK_CALL_FUNC_ARGLIST(yielded_arg, callback_arg))
     ID mid;
     const rb_callable_method_entry_t *me;
     rb_execution_context_t *ec;
+    const VALUE symbol = RARRAY_AREF(callback_arg, 0);
+    const VALUE refinements = RARRAY_AREF(callback_arg, 1);
+    VALUE klass;
 
     if (argc-- < 1) {
 	rb_raise(rb_eArgError, "no receiver given");
     }
     obj = *argv++;
-    mid = SYM2ID(callback_arg);
-    me = rb_callable_method_entry_with_refinements(CLASS_OF(obj), mid, NULL);
+
+    mid = SYM2ID(symbol);
+    for (klass = CLASS_OF(obj); klass; klass = RCLASS_SUPER(klass)) {
+        me = rb_callable_method_entry(klass, mid);
+        if (me) {
+            me = rb_resolve_refined_method_callable(refinements, me);
+            if (me) break;
+        }
+    }
+
     ec = GET_EC();
     if (!NIL_P(blockarg)) {
 	vm_passed_block_handler_set(ec, blockarg);
@@ -836,22 +878,22 @@ refine_sym_proc_call(RB_BLOCK_CALL_FUNC_ARGLIST(yielded_arg, callback_arg))
     if (!me) {
 	return method_missing(obj, mid, argc, argv, MISSING_NOENTRY);
     }
-    return vm_call0(ec, obj, mid, argc, argv, me);
+    return rb_vm_call0(ec, obj, mid, argc, argv, me);
 }
 
-static void
+static VALUE
 vm_caller_setup_arg_block(const rb_execution_context_t *ec, rb_control_frame_t *reg_cfp,
-			  struct rb_calling_info *calling, const struct rb_call_info *ci, rb_iseq_t *blockiseq, const int is_super)
+                          const struct rb_call_info *ci, const rb_iseq_t *blockiseq, const int is_super)
 {
     if (ci->flag & VM_CALL_ARGS_BLOCKARG) {
 	VALUE block_code = *(--reg_cfp->sp);
 
 	if (NIL_P(block_code)) {
-	    calling->block_handler = VM_BLOCK_HANDLER_NONE;
-	}
+            return VM_BLOCK_HANDLER_NONE;
+        }
 	else if (block_code == rb_block_param_proxy) {
-	    calling->block_handler = VM_CF_BLOCK_HANDLER(reg_cfp);
-	}
+            return VM_CF_BLOCK_HANDLER(reg_cfp);
+        }
 	else if (SYMBOL_P(block_code) && rb_method_basic_definition_p(rb_cSymbol, idTo_proc)) {
 	    const rb_cref_t *cref = vm_env_cref(reg_cfp->ep);
 	    if (cref && !NIL_P(cref->refinements)) {
@@ -859,28 +901,32 @@ vm_caller_setup_arg_block(const rb_execution_context_t *ec, rb_control_frame_t *
 		VALUE func = rb_hash_lookup(ref, block_code);
 		if (NIL_P(func)) {
 		    /* TODO: limit cached funcs */
-		    func = rb_func_proc_new(refine_sym_proc_call, block_code);
+                    VALUE callback_arg = rb_ary_tmp_new(2);
+                    rb_ary_push(callback_arg, block_code);
+                    rb_ary_push(callback_arg, ref);
+                    OBJ_FREEZE_RAW(callback_arg);
+                    func = rb_func_proc_new(refine_sym_proc_call, callback_arg);
 		    rb_hash_aset(ref, block_code, func);
 		}
 		block_code = func;
 	    }
-	    calling->block_handler = block_code;
-	}
-	else {
-	    calling->block_handler = vm_to_proc(block_code);
-	}
+            return block_code;
+        }
+        else {
+            return vm_to_proc(block_code);
+        }
     }
     else if (blockiseq != NULL) { /* likely */
 	struct rb_captured_block *captured = VM_CFP_TO_CAPTURED_BLOCK(reg_cfp);
 	captured->code.iseq = blockiseq;
-	calling->block_handler = VM_BH_FROM_ISEQ_BLOCK(captured);
+        return VM_BH_FROM_ISEQ_BLOCK(captured);
     }
     else {
 	if (is_super) {
-	    calling->block_handler = GET_BLOCK_HANDLER();
-	}
-	else {
-	    calling->block_handler = VM_BLOCK_HANDLER_NONE;
-	}
+            return GET_BLOCK_HANDLER();
+        }
+        else {
+            return VM_BLOCK_HANDLER_NONE;
+        }
     }
 }
